TAQOZ

FORGET *SDCS
( Tachyon Forth for the P2 - SD card utilities )

( 180115	3960 )
( 180111	4258 CODE BYTES )

#P61    == *SDCS            --- SDCARD CS
\ #P36    == *SDDO            --- Data out from SDCARD
\ #P41    == *SDCK            --- SDCARD clocks
\ #P40    == *SDDI            --- Data to SDCARD


{
BeMicroCV-A9            SW1
41	CLK     CLK  	60
40	CMD 	DI      58
36	D0      DO      59
37	D1
38	D2
39	D3	CS  	61
}

--- v27 remapped
( SD CARD PIN ASSIGNMENTS )
( CS.DO.DI.CK )
&61.59.58.60	== SDC	--- SW1 remapped standard SD boot pins
{
&57.40.36.41	== SDD	--- CVA9 ext SW1 remapped
&56.40.36.41	== SDE	--- CVA9 ext SW1 remapped
&15.14.13.12	== SDF		--- DE2-115 EXT
}

SDC		== =sdpins	--- Default pins
#512		== BLKSIZ
4		=: #files

            --- make the following definitions private until
--- SD Tokens
---		data token for single block read/write
$FE             =: =dtk
--- $FC         =: =dtk1            --- data token for 1st block read/write
--- $FD         =: =dtke            --- data token for last block read/write


--- SD CARD REGISTERS ---

---		Card ID
16		bytes cid
---		Card Specific Data
16		bytes csd
---		Operating conditions register
4		bytes ocr

4		bytes _sdwr
4		bytes _sdrd
---		SD CMD crc
1		bytes crc

		byte sdcmd

---		byte flag for each channel to indicate current sector buffer has been written to
#files		bytes wrflgs
---		Current sector loaded in SDBUFs for the 4 files possible
#files		longs sectors
---		Sector CRCs
#files		longs scrcs

--- constant pointers to current file variables - changed when new file channel selected via FILE
---		make @WRFLG a constant byte address, one for each file (faster than bit ops)
wrflgs		== wrflg
---		fast access to current sector (than sector @FILE)
sectors		== @sector
scrcs		== @scrc
---		initial value for SDBUF which depends upon which file handle is selected
BUFFERS		== SDBUF
---		return the current file channel
0		== FILE#





{ Detect SD card presence
the CS line must NOT have a redundant pullup on it as each
SD card has an internal pullup for this purpose
}
---				get CE pin       low,wait,float,wait,read,high
pub CARD? ( -- flg )		=sdpins 24 >> PIN L 10 us F 0 DROP R H ;

---				dummy clocks (enables CE)
pri SDCLK			-1 SPIWR DROP ;

pre ;;				@CODES @ 2- W++  ;

--- do what needs to be done if the SD card is busy (LED etc) - can be rewritten
{ NOT NEEDED FOR ROM VERSION
pub SDRDY			NOP ;
pub SDBSY			NOP ;
pub SDERR			NOP ;
pub SDOK			NOP ;
}

--- set pins, save them for reboot and init with !SDIO see high level !SD
pub SDPINS ( ce-miso-mosi-clk -- )
	--- remember which pins were used for BACKUP
	DUP			' =sdpins ==!
    	DUP >B			' *SDCK ==!
	--- setup SDCS mask
	#24 >>			' *SDCS ==!
	;


( SD CARD RESPONSES )


--- Wait for read token and read SD data into buffer
pri TOKEN ( marker -- flgX )
---	timeout             token?     no, countdown and terminate?
	#20,000 BEGIN OVER 0 SPIRD <> WHILE 1- DUP 0= IF NIP EXIT THEN REPEAT
	2DROP TRUE
	;



--- Send the command to the SD card and read result
pub CMD    ( data cmd -- res )
	DUP sdcmd C!
---	ensure SPI instruction is setup for SD card pins
	=sdpins SPIPINS
---	preclock & write command
	SDCLK 	  SPIWC DROP
---	write 4 bytes data
	SPIWR SPIWR SPIWR SPIWR DROP
---	write crc (don't care except during init)
	crc C@ SPIWB DROP
---	read card result
pri SDRES ( -- res )
	0 BEGIN SPIRD >B DUP $FF <> UNTIL
	;


--- Send an ACMD to the card and return with response
pub ACMD ( data acmd -- res )
	0 #55 CMD DROP CMD
	;;
pri STAT@ ( -- stat )
	0 #13 CMD SDRES 8<< OR
	;

pri SDERR? ( -- flg ; return SD bit flag errors)
	0 CARD? 0= 1 AND OR	--- b0 = card detect error
	_sdrd @ 0= 2 AND OR	--- b1 = read failure
	_sdwr @ 0= 4 AND OR	--- b2 = write failure
	ocr @ 0= 8 AND OR	--- b3 = card error
	;


pri SD4@ ( -- long )			0 SPIRD SPIRD SPIRD SPIRD ;

--- Wait for read token and read SD data into buffer
pri SDDAT! ( adr -- )
	=dtk TOKEN IF 16 ADO 0 SPIRD I C! LOOP SDCLK SDCLK SDCLK ELSE DROP THEN
	;

pri !SDX
	sdcmd C~~
---	Init into SPI mode with retries ( Sandisk 8G takes 3 retries)
	5 FOR $95 crc C! 0 0 CMD 1 = ?NEXT 0EXIT
	5 FOR $87 crc C! $1AA 8 CMD 1 = ?NEXT 0EXIT
---	check pattern correct
        SD4@ $1AA = 0EXIT
--- Initialization R1 b30=HCS (requires quite a bit of time vs just clocks)
\	 1000 FOR SDCLK $4000.0000 #41 ACMD 0= ?NEXT 0EXIT
	0 1000 0 DO
	   $4000.0000 #41 ACMD
---	     100 slow clocks with chip disabled
	     IF SPICE ( *SDDI HIGH ) =sdpins C@ PIN 200 FOR H L NEXT
	     ELSE 1+ LEAVE
	     THEN
	LOOP
	0EXIT
	--- read OCR R3  - OCR(30)=CCS
	0 #58 CMD ?EXIT SD4@ DUP ocr !
	IF
	--- optional read cid and csd
	  0 #10 CMD 0= IF cid SDDAT! THEN
	  0 9 CMD 0= IF csd SDDAT! THEN
	THEN
	;

--- Initialise the SD card in SPI mode and return with the OCR
pub !SD ( -- ocr|false )
	=sdpins	SPIPINS
\	SDBSY
\	SDOK
	--- erase all module variables, selects file
	cid sdcmd 1+ OVER - ERASE
	--- erase all n file buffers
	SDBUF #files 9<<  ERASE
	--- force an invalid sector (nothing cached)
	sectors #files 4* -1 FILL
	--- stop slow clock from smartpin
	CARD?
	IF 20 FOR !SDX sdcmd C@ 9 = ?NEXT DROP THEN
\	SDRDY
	ocr @
---	indicate an error visually if enabled
\	DUP NOT IF SDERR THEN
	;



{ *** BLOCK BUFFERS *** }


{ HELP SDWR ( src sect -- flg )
Write to a sector in the SD card - measured 2.84ms on 4GB SanDisk
Testing 32k write: LAP 0 FILE@ 9 >> #64 ADO DUP I SDWR DROP BLKSIZ + LOOP LAP .LAP 274.111ms ok
 TF2: LAP 0 FILE@ 9 >> #64 ADO DUP I SDWR DROP BLKSIZ + LOOP LAP .LAP 88.418ms
 TF2: LAP 0 FILE@ 9 >> #64 ADO DUP I SDWR DROP BLKSIZ + LOOP LAP .LAP 74ms (using SPITX)
 TF2 single sector = 1.275ms
}
pub SDWR    ( src sect -- flg ; Write from src to xdst in the SD )
\	SDBSY
---	Prep card
	SDCLK SDCLK SDCLK
---	write block command
	#24 CMD 0= ( src res )
	IF
	  SDCLK SDCLK SDCLK =dtk SPIWB DROP
---	  write block to SD
	  BLKSIZ SPITX
---	  wait for return token
	  0 TOKEN $FF TOKEN AND
	ELSE FALSE
	THEN
	SPICE
\	SDRDY
---	save write flag
	DUP _sdwr !
	;

{
card has been prep'd for read  - proceed and read a block of data
write the current sector back to the storage media
read crc and force as true flag
}
pri SDRDBK ( dst -- crc|flg )	BLKSIZ SPIRX 0 SPIRD SPIRD 1 OR ;

--- Write the sector buffer if it has been modified or force (old WRSECT)
pub FLUSH ( force -- )		wrflg C@ OR IF SDBUF @sector @ SDWR DROP wrflg C~ THEN ;


{ HELP: SECTOR ( sect -- )
Read the sector into the SDBUF
takes around 1.9ms, 2.55ms
}
pub SECTOR ( sect -- )
---	skip it if we already have  its
	DUP @sector @ = IF DROP EXIT THEN
---	update sector pointer for this file channel
	0 FLUSH DUP @sector !
	SDBUF

--- SD CARD READ BLOCK - Read sector from SD into dst
pub SDRD    ( sector dst --  )
\	SDBSY
---	read block command
	SWAP #17 CMD ( data cmd -- res )
	IF
---	  didn't like that command - discard dst & return false
	  DROP FALSE
	ELSE
---	  wait for data - read block else terminate early & return false
	  SDRES =dtk = IF SDRDBK ELSE DROP SDCLK SDCLK STAT@ DROP FALSE THEN
	THEN
\	SDRDY
---	extra clk and save crc/flg
	SDCLK DUP _sdrd ! @scrc !
	SPICE
	;

--- Read multiblock sectors from SD into dst
pub SDRDS    ( sector dst sectors -- crc | false )
\	SDBSY
	-ROT
	SWAP #18 CMD ( data cmd -- res )			--- read block command
	IF
	  --- didn't like that command - discard dst & return false
	  2DROP FALSE
	ELSE
	  --- process read token and read block if available
	  SDRES =dtk = IF SWAP FOR DUP SDRDBK DROP BLKSIZ + NEXT ELSE 2DROP SDCLK SDCLK STAT@ DROP SPICE FALSE THEN
	THEN
	--- release the SD card
\	SDRDY
	SDCLK SPICE
	---  save crc/flg
	DUP _sdrd !
	BEGIN 0 12 CMD 0= UNTIL
	SPICE
	;



--- --- --- --- --- --- --- ------ --- --- ------ ------ --- --- ------ --- --- ---



( Virtual memory addressing is suitable for addressing the first 4GB of the memory )


\ 0	== @SECTOR

--- Translate a virtual address into a HUB RAM address - may have to flush and read a sector

pri WRSDA ( xaddr -- addr )	wrflg C~~
--- translate absolute 4GB SD memory address to a buffered address in hub RAM
pub RDSDA ( xaddr -- addr )	L>S ( @SECTOR + ) SECTOR SDBUF + ;


--- read and write from virtual memory address xaddr
pub SD@ ( xaddr -- long )	RDSDA @ ;
pub SD! ( long xaddr -- )	WRSDA ! ;

pub SDC@ ( xaddr -- byte )	RDSDA C@ ;
pub SDC! ( byte xaddr -- )	WRSDA C! ;

pub SDW@ ( xaddr -- word )	RDSDA W@ ;


--- DUMP MEMORY ACCESS METHOD MODIFIER
--- Usage: 0 $200 SD DUMP
pub SD				DUMP: SDC@ SDW@ SD@ ;


#files		longs frd
#files		longs fwr

--- Set the active file channel (0 ..n) by setting psuedo constants
pub FILE ( index:0..n -- )
     #files //
     DUP ' FILE# ==!
     DUP 4* sectors + ' @sector ==!
     DUP 4* scrcs + ' @scrc ==!
     DUP wrflgs + ' wrflg ==!
     9<< BUFFERS + ' SDBUF ==!
     ;;



--- access via @FILE for current file

pub @FRD ( -- readptr )			frd
{ HELP @FILE ( addr -- addr+off )
returns the address offset into sector & sectcrc tables for active file channel
as set with    index FILE
e.g. val sector @FILE ! otherval sectcrc @FILE ! sector @FILE @
}
pub @FILE ( addr -- addr+off )		FILE# 4* +  ;

--- access via @FILE for current file

pub @FWR ( -- writeptr )		fwr @FILE ;;
pub FWR@				@FWR @ ;








--- PARTITION TABLE ---

--- Partition table image - copied here
16	longs parts             --- Room for 4 entries of 16 bytes

--- parts 8 +	== fatptr	--- long that points to sector containing fat info
pri @INFO		parts 8 + @ ;

--- These are the offset constants compiled as absolute addresses (for 1st partition)
--- Many names have not been added to the dictionary to save space (but could be reclaimed later)
{

Offset 	Description 	Size
00h 	Current State of Partition(00h=Inactive, 80h=Active) 	1 Byte
01h 	Beginning of Partition - Head 	1 Byte
02h 	Beginning of Partition - Cylinder/Sector (See Below) 	1 Word
04h 	Type of Partition (See List Below) 	1 Byte
05h 	End of Partition - Head 	1 Byte
06h 	End of Partition - Cylinder/Sector 	1 Word
08h 	Number of Sectors Betweenthe MBR and the First Sector in the Partition 	1 Double Word
0Ch 	Number of Sectors in thePartition 	1 Double Wor

parts    ORG
1    res --- state		--- 00 = inactive, $80 = active
1    res --- head		--- beginning of paritiion
2    res --- cylsec		--- beginning of partition cylinders/sector

1    res --- partype		--- type of partition
1    res --- ehead		--- end of partition head
2    res --- ecylsec		--- end of partitiion cylinder/sector

4    DS fatptr
4    res --- size
pub .parts
	4 0 DO I 4 SHL $1BE +
	CR DUP C@ .byte SPACE
	DUP 1+ C@
	;

File System Information Sector
Offset 	Description 	Size
00h 	First Signature (52h 52h 61h41h) 	1 Double Word
04h 	Unknown, Currently (Mightjust be Null) 	480 Bytes
1E4h 	Signature of FSInfo Sector(72h 72h 41h 61h) 	1 Double Word
1E8h 	Number of Free Clusters (Setto -1 if Unknown) 	1 Double Word
1ECh 	Cluster Number of Clusterthat was Most Recently Allocated. 	1 Double Word
1F0h 	Reserved 	12 Bytes
1FCh 	Unknown or Null 	2 Bytes
1FEh 	Boot Record Signature (55hAAh) 	2 Bytes

FAT32 Drive Layout
Offset 	Description
Start of Partition 	Boot Sector
Start + # of ReservedSectors 	Fat Tables
Start + # of Reserved + (#of Sectors Per FAT * 2) <Assuming that FAT Mirroring isEnabled, I personally haven't seen a case where itwasn't, but I guess there is always the possibility> 	Data Area (Starts withCluster #2)


}

--- BOOT RECORD ---

{
Boot Record
The boot record occupies one sector, and is always placed in logical sector number zero of the "partition". If the media is not divided into partitions, then this is the beginning of the media. This is the easiest sector on the partition for the computer to locate when it is loaded. If the storage media is partitioned (such as a hard disk), then the beginning of the actual media contains an MBR (x86) or other form of partition information. In this case each partition's first sector holds a Volume Boot Record.

vbr
40.0000: EB 3C 90 4D 53 57 49 4E 34 2E 31 00 02 40 0E 11    .<.MSWIN4.1..@..
40.0010: 02 00 00 00 00 F8 00 00 3F 00 FF 00 00 20 00 00    ........?.... ..
40.0020: C0 37 EF 00 79 07 00 00 00 00 00 00 02 00 00 00    .7..y...........
40.0030: 01 00 06 00 00 00 00 00 00 00 00 00 00 00 00 00    ................
40.0040: 80 01 29 87 D6 12 00 50 32 53 44 43 41 52 44 20    ..)....P2SDCARD
40.0050: 20 20 46 41 54 33 32 20 20 20 00 00 00 00 00 00      FAT32   ......

problem vbr 16GB
00.E13E: 45 46 49 20 50 41 52 54 00 00 01 00 5C 00 00 00    EFI PART....\...
00.E14E: 89 B1 D7 CB 00 00 00 00 01 00 00 00 00 00 00 00    ................
00.E15E: FF CB DA 01 00 00 00 00 22 00 00 00 00 00 00 00    ........".......
00.E16E: DE CB DA 01 00 00 00 00 D9 F8 06 10 B2 F6 8D 44    ...............D
00.E17E: 81 B9 CE A0 1F 9D 89 E5 02 00 00 00 00 00 00 00    ................
00.E18E: 80 00 00 00 80 00 00 00 DE 97 FF 00 FF FF FF FF    ................
00.E19E: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00    ................
00.E1AE: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00    ................ ok

fsis
40.0200: 52 52 61 41 00 00 00 00 00 00 00 00 00 00 00 00    RRaA............
40.03E0: 00 00 00 00 72 72 41 61 1F 77 00 00 43 2A 01 00    ....rrAa.w..C*..
40.03F0: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 55 AA    ..............U. ok

}

90	bytes vbr		--- 96 byte array
{
3	res
8	res 			--- 3: oemname
2	res  			--- 11: byte/sect
1	res			--- 13: sectors/cluster
2	res  			--- 14: rsvd
--- $10 ---
1	res  			--- 16: Copies of FAT
4	res			--- root entry count (FAT16)
1	res			--- media $F8
2	res			--- n/a
2	res			--- sect/trk
2	res			--- heads
4	res			--- hidden sectors
--- $20 ---
\ 4	bytes capacity		--- 32: Number of sectors * byte/sect (512) = capacity
4	res
--- $24 ---
4	res   			--- 36: Number of sectors per FAT table
2	res			--- 40: fatflags
2	res			--- 42: fatver
4	res 			--- 44: Cluster Number of the Start of the Root Directory (02 -> $4000)
\ 2	bytes fsis		--- info = Sector Number of the FileSystem Information Sector  (from part start)
2	res 			--- filesystem information sector
2	res			--- boot = Sector Number of the Backup Boot Sector (from part start)
#12	res
--- $40 ---
3	res
4	res 			--- #67 serial number of partition
#11	res 			--- #71 volume name
8	res 			--- #82 always FAT32 - (don't trust)
--- END $5A ---
}



--- create room for some system variables in this table
1	bytes mounted		--- true flag if mounted (but also depends upon other checks)
\ 4	bytes rootdir		--- sector address of root directory


--- DIR BUFFERS ---

--- A copy of the original virtual address of the directory entry is kept in this array (4 entries)

4	longs diradr		--- virtual memory address of file's directory entry

4	longs _file		--- table entry for 4 open files - holds sector address

--- Directory entry table
32	longs dirbuf		--- create a directory buffer for the opened file

	byte _rw		--- file r/w permissions

	long @cwd
	long fboot		--- boot signature - determines whether it needs to remount


--- index into volume boot record
pri vbr+ ( index -- adr ) 	vbr + ;

--- file read/write modes
pri @RW		FILE# |< _rw ;
pub RW?		@RW SET? ;
pub RW		@RW SET ;
pub RO		@RW CLR ;

--- FAT HANDLERS ---
---					long pseudo-constant
pri @ROOT ( -- rootdirsect )		$FFFF.FFFF ;
pub ROOT 				@ROOT @cwd ! ;

--- return with the starting sector of the selected FAT (normally 0 or 1)
pri @FAT ( fat# -- sector )
	36 vbr+ @ *
	@INFO 14 vbr+ W@ + +
	;

--- FAT pointers (changed by RDFAT)
0    == FAT1
0    == FAT2

--- convert cluster to sector		*--- changed by RDFAT
pri SC* ( cluster -- sector )		6 << ;
--- sectors/cluster
pri S/C ( -- sectors )			#13 vbr+ C@ ;

pri RDFAT
	@INFO SECTOR					--- Read the FAT32 boot record
	SDBUF vbr #90 CMOVE				--- backup the FAT32 entry for direct access
	S/C >| ' SC* C!					--- make a fast multiply using a shift left constant
	#14 vbr+ W@ #36 vbr+ @ #16 vbr+ C@ *
	#44 vbr+ @ 2- SC* + + @INFO + ' @ROOT ==!	--- rootdir constant
	0 @FAT ' FAT1 ==!				--- save time by precalculating FAT table addresses
	1 @FAT ' FAT2 ==!
	;;

\ HERE 6 +	== drive$	0 , 0 ,
\ 8 bytes drive$


pri .SER		#67 vbr+ @ .LONG SPACE ;;
pri .VOL		#71 vbr+ #11 CTYPE SPACE ;;

\ pri .CID		"[" EMIT cid 1+ 7 ADO IC@ $20 MAX $7E MIN EMIT LOOP "]" EMIT SPACE ;
pri .CID		cid 1+ 7
pub .[ ( str cnt -- )	'[' EMIT CTYPE ']' EMIT SPACE ;;

pri .FAT
	CRLF ." Mounted " ( drive$ PRINT$ SPACE )
	--- read FAT32 boot record
	@INFO SECTOR .VOL
	--- display card & fat serial
	cid 9 + @ .long '-' EMIT .SER
	--- card info name & oemname
	.CID 3 vbr+ 8 CTYPE SPACE
	--- FAT32 names
	#82 vbr+ 8 CTYPE SPACE
	--- display memory capacity
	#32 vbr+ @ BLKSIZ 1,000,000 */ .DECL ." MB ("
	--- display cluster size
	S/C 2/ PRINT ." kB/cluster) "
	;



pub UNMOUNT		parts fboot OVER - ERASE ;

\ pub DRIVE ( name$ -- )	drive$ $!
{ HELP MOUNT
Mount the currently selected storage device and init all 4 file handles
Read the FAT32 and set variables accordingly
}
pub MOUNT
\	SDERR					--- turn on error LED - only cleared if FAT passes muster
	UNMOUNT !SD
	IF
	  mounted C~~
	  --- read and check partition info
	  0 FILE $01FE SDW@ $AA55 =
	  IF
	    --- buffer patitions
	    $01BE RDSDA parts $40 CMOVE
	    --- read the FAT32 sector and display
	    RDFAT .FAT
	    --- use root directory
	    ROOT
	    --- clear error indication & copy boot signature to validate mount status
\	    SDOK
	  BOOT? fboot !	\  bootsig @ fboot !
	  ELSE
            ."  *MBR Format Error* "
	  THEN
	ELSE
	  ." *Card Error* "
	THEN
	;

pub ?MOUNT
    CARD? NOT
     IF
       ."  *No Card inserted!* "
\       SDERR
	R> DROP
       fboot ~
     THEN
     fboot @ BOOT? <> IF MOUNT THEN
\    bootsig @ fboot @ <> mounted C@ 0= OR IF MOUNT THEN

     ;


\ pub C:        SDC SDPINS " C:" DRIVE ;
{
pub D:        SDD SDPINS " D:" DRIVE ;
pub E:        SDE SDPINS " E:" DRIVE ;
pub F:        SDF SDPINS " F:" DRIVE ;
}

{
pri ?NEW
	CARD? NOT
	  IF mounted C~ EXIT
	  ELSE mounted C@ 0= IF 50 ms MOUNT THEN
	  THEN
	;



1	bytes _card		--- card detect transition memory

--- Make sure the card is inserted or mount it if it has just been inserted

pub ?SDCARD
    _card C@ 0= CARD? DUP _card C! AND			--- check previous card state with now and update
      IF MOUNT THEN					--- just inserted so mount automatically
    CARD? NOT IF ( SDERR ) mounted C~ THEN		--- indicate an error if it's not inserted
    ;


--- Init file system
pub !FS				mounted C~ ' ?NEW keypoll W! ;

}








--- FILE NAME HANDLERS ---

16 #files *	bytes _file$
16		bytes fname$


\ " FILENAME.TXT" 16 4 * STRING file$			--- The actual name of the file requested
--- file$ stores 4 8.3 filenames at 16 byte boundaries

pub FILE$  ( -- addrOfFilenameString )    FILE# 4* 4* _file$ + ;


--- format friendly file name into directory format
--- FILE.TXT --> FILE    TXT
pri >F83 ( str1 -- str2 )
---	prep fname$ as all blanks with terminator
	fname$ #11 $20 FILL
---	ensure terminator is set
	fname$ #11 + C~
	fname$ BEGIN OVER C@ DUP '.' <> SWAP 0<> AND WHILE OVER C@ OVER C! 1+ SWAP 1+ SWAP REPEAT
	--- process the extension if it has one
	DROP DUPC@ IF 1+ DUP LEN$ ?DUP IF fname$ 8 + SWAP CMOVE ELSE 2DROP THEN ELSE DROP THEN
	fname$
	;


--- DIRECTORY STRUCTURE ---

\ #00    == @FNAME      --- File name
\ #08    == @FEXT       --- File extension
#11    =: @ATR		--- Attribute
\ #12    == @CASE	--- reserved - used for case info
\ #13    == @CMS        --- creation millisecond (2ms)
\ #14    =: @CTIME      --- creation time
\ #16    =: @CDATE      --- creation date
\ #18    =: @ADATE      --- access data
\ #20    == @FCLSTH	--- First cluster (high word)
#22    =: @FTIME        --- modification time
#24    =: @FDATE        --- modification date
\ #26    == @FCLST      --- First cluster of file (low word)
#28    =: @FSIZE        --- size of file


pri SECT1 ( diradr -- sector )
--- = virtual memory address of file's directory entry
	DUP #26 + W@ SWAP #20 + W@ W>L
\ pub CLUST>SECT ( clust# -- sector )
	#44 vbr+ @ - SC* @ROOT +
	;

--- return with the address of the 32-byte directory buffer  in RAM using current file index
pub @DIRBUF ( -- bufadr )	FILE# 5 << dirbuf + ;


#12 bytes dir$
pub DIR?    ( str -- diraddr>>5 | FALSE )
	?MOUNT
	dir$ #11 + C~
	mounted C@
	IF
	  0 ( str offset )
	  BEGIN ( str dirptr )
	    DUP 9 >> @cwd @ + SECTOR			--- ensure correct sector is loaded
	    DUP 9BITS SDBUF + DUPC@			--- terminate on null entry
	    3RD 0= OR					--- but not first one
	  WHILE ( str index bufptr )
	    DUP $E5 <>					--- skip deleted entries
	    IF
	    DUP @DIRBUF #32 CMOVE			--- save in case it's successful so FOPEN# can access
	    dir$ #11 CMOVE				--- copy the name as a string
	    dir$ TRUE 4TH 11
	      ADO OVER C@ IC@ = AND SWAP 1+ SWAP LOOP
	    NIP
	      IF NIP 5 >> @cwd @ 4 << +  EXIT THEN	--- Found it, exit here with address
	    THEN
	    $20 +					--- skip to next dir entry (32 bytes)
	  REPEAT
	  DROP 2DROP					--- failed to find it - discard parameters
	ELSE
	  DROP
	THEN
	FALSE						--- Failed to find it
	;


--- DIRECTORY ENTRY HANDLERS ---

--- update current directory entry from buffer
{
pub DIR!
	@DIRBUF diradr @FILE @ RDSDA $20 CMOVE		--- copy that to the virtual address
	WRSECT DROP					--- Force an update
	;

pub DIR!
---	diradr holds 64GB address>>5 for 32-byte aligned directory entries
	@DIRBUF diradr @FILE @
	DUP 4 >> SECTOR >N 5 << SDBUF +
---	copy that to the virtual address & force an update
	$20 CMOVE  ( non-zero flag ) FLUSH
	;




--- split 6 digit decimal number into 3 two digit groups
pri DECS ( #xxyyzz -- zz yy xx )	#100 U// #100 U// ;

pre ENTER	HERE 2+ [C] || ;

pri DECS ( #xxyyzz -- zz yy xx )
	TWICE
	#100 U//
	;



--- write to directory entry as new date
pri +DIR! ( word field -- )		@DIRBUF + W! ;

--- update file modification/create date in dir buf
pri FDATE! ( #yymmdd field -- )                --- Date (7/4/5 bits, for year-since-1980/month/day)
	SWAP
	#200000 +                     --- arrange as decimal YYMMDD from 1980 ( 2000.0000 + 1980.0000 - )
	DECS 9<< SWAP 5 << + +
	SWAP +DIR!
	;

--- update file modification/create time in dir buf
pri FTIME! ( #hhmmss field -- )                 --- Time (5/6/5 bits, for hour/minutes/doubleseconds)
	SWAP DECS #11 << SWAP 5 << + SWAP 2/ +
	SWAP +DIR!
	;

--- DATE TIME STAMPING ---

#150101    == DATE@
#120000    == TIME@


--- Update the modified time and data of the current file
pub FSTAMP ( -- )		DATE@ @FDATE FDATE! TIME@ @FTIME FTIME! DIR! ;

}
pri .FYEAR ( fdate -- )		9>> #1980 + .AS"  #### " ;

pri .ASMONTH ( index -- )	>N 1- 3 * " JanFebMarAprMayJunJulAugSepOctNovDec" + 3 CTYPE ;;
--- print date in Unix format
--- 	extract day of the month to index into name string
pri .FDATE ( fdate -- )		DUP 5 >> .ASMONTH $1F AND  .AS"  #`#  " ;

pri .FTIME ( ftime -- )		DUP #11 >> 100 W* SWAP 5 >> $3F AND OR .AS"  ##:## " ;



--- CLUSTER HANDLERS ---

{
--- display cluster usage from FAT
pub .FATMAP
     36 vbr+ @ 0 DO
       I $3F AND 0= IF CR I .WORD SPACE  THEN
       FAT1 I+ SECTOR
       sectcrc @FILE W@
       IF '*' ELSE '.' THEN EMIT
       char C@ $1B = IF LEAVE THEN
     LOOP
     ;

}
{
pri NextFreeDir ( -- xadr )
     0
     cwdsect @ 9<< S/C 9<<
     ADO I SDC@ 0= IF I+ LEAVE THEN 32 +LOOP
     ;
}

{ CLUSTER CHAIN CODES
If value => $0FFF.FFF8 then there are no more clusters in this chain.
$0FFF.FFF7 = bad
0 = free

}
pri @CLUST ( index -- xadr )     4* FAT1 9<< + ;
{
pri CLUSTER@ ( index -- cluster )
      @CLUST SD@
     ;
}

{ HELP CLUSTERS? ( size -- xadr )
Find free clusters for the file size in bytes - 0 = all
return with address of first free cluster
}
{
pri CLUSTERS? ( size -- xadr )
--- calculate clusters required
     9>> S/C / ( clusters )
      0
     BEGIN
---	find a free cluster
       BEGIN DUP @CLUST SD@ WHILE 1+ REPEAT
       ( clusters index )
---	check for sufficient contiguous clusters (linear non-fragmented)
       0 OVER 4TH ADO I @CLUST SD@ OR DUP IF NIP I SWAP LEAVE THEN LOOP
       ( clusters chain flag )
     WHILE
       1+
    REPEAT
     NIP
     ;
}

pub FSIZE@				@DIRBUF @FSIZE + @ ;

pub FSECT@ ( -- sect )			_file @FILE @ ;


--- VIRTUAL FILE ACCESS ---

pri FSA! ( faddr -- addr )		wrflg C~~
pub FSA ( faddr -- addr )		L>S FSECT@ + SECTOR SDBUF + ;


--- common FS store routines
pri FSX! ( faddr -- addr )		FSECT@ IF FSIZE@ UMIN FSA! ELSE R> 3DROP THEN ;

pub FSC! ( byte faddr -- )		FSX! C! ;

pub FS! ( long faddr -- )		FSX! ! ;

--- 17.212us
pub FSC@ ( faddr --- byte )		FSA C@ ;

pub FSW@				FSA W@ ;

pub FS@					FSA @ ;

--- Redirect next dump to use file system memory
pub FS					DUMP: FSC@ FSW@ FS@ ;


--- SEQUENTIAL FILE ACCESS ---

---		current status of file system ( pseudo-constant)
pub FS?		0 ;
pub FABORT ( ch -- )			' FS? C! R> DROP ;


--- return with remaining bytes in file
pub FREM ( --- rem )			FWR@ DUP IF FSIZE@ SWAP - THEN ;


word fkey						--- backup for input device when input is switched to file

--- Read in the next character in the virtual buffer as part of the console input stream
pri FGET ( -- ch )
---	Read a single character
	@FRD @ FSC@
---	stop when null encountered and return to previous input device
	DUP 0= IF fkey W@ ukey W! 'X' FABORT THEN
---	update index otherwise
	@FRD ++
	;

--- Write a character into the logical end of the file and update the write pointer
pub FPUT ( ch -- )
	?DUP 0EXIT
							--- ignore null characters (but will write nulls if another bit in long is set)
pub FPUTB ( byte -- )					--- write byte even if it is null
	FSECT@ 0<> RW? AND					--- check if file is opened and a valid write pointer is set (APPEND)
	IF							--- don't write if a file is not opened
	  FWR@ FSIZE@ BLKSIZ ALIGN =>			--- file full?
	  IF DROP 'O' FABORT THEN			--- Don't write to beyond end of file - FABORT = O for OVERFLOW
							--- still room in the file so just write it
	FWR@ FSC!					--- write the character to the file
	FWR@ FSIZE@ BLKSIZ ALIGN  1- =			--- last byte in file?
	IF 'F' FABORT THEN				--- Warn file is now full - FABORT = F for FULL, exit now!
	  @FWR ++
	  0 FWR@ FSC! ;;					--- Insert a null terminator (overwritten by the next character)
	THEN
	DROP 'N' FABORT					--- no file - FABORT = N
	[



\ TAQOZ# $4.0000 $1.0000 LAP FREAD LAP .LAP 5911280 cycles = 73.891ms  ok
\ SPI>BUF w CALL	 $2.0000 256 KB LAP FREAD LAP .LAP 21455057 cycles = 268.188ms  ok
\ SPI>BUF inc SPIRD	$2.0000 256 KB LAP FREAD LAP .LAP 20415696 cycles = 255.196ms  ok
--- File Read to memory
pub FREAD ( dst size --- )	9>> FSECT@ -ROT SDRDS DROP ;





{ HELP >FILE ( -- )
Redirect character output via uemit to the open file using "fptr" which is set to the start of the file when opened
If the file is not opened and a valid write pointer set then output will be discarded
}

pub >FILE ( -- )	FSECT@ IF ' FPUT uemit W! ELSE NONE THEN ;

---			flush   reset sect   clear name           wipe dir entry
pub FCLOSE		0 FLUSH	@sector ~~ _file @FILE ~ FILE$ C~ @DIRBUF $20 ERASE ;


{ HELP FOPEN# ( diradr -- sector )
Open the file pointed to by the 64GB virtual directory entry address
aligned to 32 byte directory entries
}
pri FOPENX ( xdiradr5 --- sector )
	diradr @FILE !					--- save 64GB address of directory entry
	@DIRBUF	SECT1
	_file @FILE !					--- save starting sector of file
	@FRD ~	@FWR ~
	RO FSECT@
	;;

{ HELP FOPEN$ ( namestr -- sector )
Open the file with the 8.3 name and return with its sector or 0 if failed
The variable FABORT can be checked for more information if there was an error
Execution time ~ 10ms sample:
" FIRMWARE.ROM" LAP FOPEN$ DROP LAP .LAP 2.887ms ok
Usage: " SYSLOG.TXT" FOPEN$ IF etc etc
}
pub FOPEN$ ( namestr -- sector )
---  skip blanks name
     FCLOSE DUPC@
     IF
       DUPC@ '/' = IF 1+ THEN				--- ignore leading / in file name
       DUP LEN$ #12 >
         IF  DROP FALSE '!' FABORT THEN		--- abort with ! status if name is too long
       DUP FILE$ $!					--- set the name for this file
      >F83 DIR? DUP					--- find the dir entry
        IF FOPENX ;; THEN    ( sector )				--- open the file from the dir entry
     THEN
     DROP FALSE '?' FABORT
    [

{ HELP FOPEN ( <name> -- )
Open a file interactively with the name specified in the input stream
Report back to the console
}
pre FOPEN ( <name> -- )
    GET$ FOPEN$ ?DUP
       IF ." ...opened at "  .long SPACE
       ELSE ."  not found "
       THEN
    ;


{
--- FILE SHELL COMMANDS ---
--- update local copy of directory entry
pub RENAME$ ( new$ -- )		FSECT@ IF >F83 @DIRBUF #11 CMOVE DIR! THEN DROP	;

pre RENAME ( <oldname> <newname> )
    GET$ FOPEN$ IF ."  to " GET$ RENAME$ ELSE DROP ."  No file opened " THEN
    ;
}

\ pre DELETE ( <name> -- )	GET$ FOPEN$ IF  THEN ;

--- Load a file as console input - scripts or source code
pre FLOAD ( <name> -- )		GET$ FOPEN$ 0EXIT
--- Redirect the console input to read the file sequentially
pub FILE>			FSECT@ IF ukey W@ fkey W! ' FGET ukey W! THEN ;

--- exception handler - if word not found then run from file - point unum to this code
pub FRUN			@WORD DUPC@ '.' = IF 1+ THEN FOPEN$ 0EXIT FILE> ;;


pre cat ( <name> -- )
	[C] FOPEN CRLF FSECT@ 0EXIT
---  clear any preexisting console escape
	char C~ 0
	BEGIN
---    null end of file or escape key?
	  DUP FSC@ DUP char C@ $1B <> AND
	WHILE
--- make sure all LFs are preceded by a CR
	  DUP $0A = IF CR THEN
	  EMIT 1+
	REPEAT
	2DROP
	;


--- DIRECTORY PATH CONTROL ---

16 bytes cwd$



--- change directory
pre cd
	GET$
\ pub cd$ ( dirstr -- )
     --- from root directory?
	DUPC@ '/' = IF ROOT cwd$ C~ 1+ THEN
	DUP FOPEN$
	IF
	  $10 @DIRBUF @ATR + SET?
	  IF cwd$ $! FSECT@ @cwd ! EXIT THEN
	THEN
	DROP
	;

pub pwd			"/" EMIT cwd$ PRINT$ ;;

\ pub .PATH		." TAQOZ@" .SER pwd ."  $ " ;


{ HELP PROMPT ( on/off -- )
Turn on the path prompt and and suppress the ok response or revert back to normal Forth ok
}

\ pub PROMPT ( on/off -- )        DUP NOT OK ' .PATH AND prompt W! ;




--- DIRECTORY LISTING ---

{
DIRECTORY ENTRY

Directory Attribute flags
0    Read Only
1    Hidden
2    System
3    Volume label
4    Directory
5    Archive

Directory types - first byte
Normal record with short filename - Attrib is normal
long filename text - Attrib has all four type bits set
Unused - First byte is 0xE5
End of directory - First byte is zero
}

{
--- print the symbol(s) for each active directory name attribute
pri .ATR ( atr -- )
    " rhsvda" SWAP 6
     FOR
       DUP 1 AND IF OVER C@ EMIT ELSE ." ." THEN
       2/ SWAP 1+ SWAP
     NEXT 2DROP
    ;
}
pri .ATR ( atr -- )
    " rhsvda" 6
     FOR
       OVER 1 AND IF DUPC@ EMIT ELSE DOT THEN
       SWAP 2/ SWAP 1+
     NEXT 2DROP
    ;

{ Formatting of file size------------------------------****
SEE     .FTH .....a 0002.34C0    Jul  6   2017  01:26  2187
SPLAT-V4.FTH .....a 0002.3500    Mar  9   2017  12:14   18k
WARPEACE.TXT .....a 0002.3540    Aug 30   2015  07:03    3M
LOVE    .WAV .....a 0002.4E00    Feb 16   2015  08:06   14M
POPCORN .WAV .....a 0002.BDC0    Jun 17   2014  06:07  117k
CD      .ISO .....a 0002.BEC0    Aug 23   2013  11:19  4.0G
}
pri .FSIZE ( diradr -- )
	@FSIZE + @
	SPACE
	DUP 10,000 U< IF .DEC4 ;; THEN
	1,000 U/
	DUP 1,000 < IF .AS" ##`#k" EXIT THEN
	1,000 U/
	DUP 1,000 < IF .AS" ##`#M" EXIT THEN
	100 U/          .AS" #.#G"
	;


( FIRMWARE.ROM .....a 0000.620A    07/07/2014 11:58:14   5,536 )

pri .F83	DUP 8 CTYPE DOT 8 + 3 CTYPE ;;

pri (.DIR)
	CRLF I
pri .DIR ( addr -- )
     --- ignore blank entries as closed
	DUPC@
	IF
---	print the file name with spaces included (formatted listing)
	  DUP .F83
--- 	print attributes as alphas
	  SPACE DUP @ATR + C@ .ATR SPACE
--- 	print first sector number
	  DUP SECT1 .long
	  4 SPACES DUP @FDATE + W@ DUP .FDATE .FYEAR
	  DUP @FTIME + W@ .FTIME .FSIZE
	ELSE
	  DROP ." *CLOSED*"
        THEN
	;

{
TAQOZ# DIR
Mounted NO NAME     7803.AA00-F84E.1690 [  SDC  ]          FAT32    4,119MB (32kB/cluster)
/NO NAME
WARPEACE.TXT .....a 0000.4040    Aug 30   2015  17:27    3M
EXTEND  .FTH .....a 0000.5900    Jul 24   2015  12:13  112k
}

pri DODIR	NOP ;

--- use this method for listing the directory
pub DIR		' (.DIR)
pri (DIR) ( cfa -- )
	?MOUNT CRLF ( drive$ PRINT$ ) pwd .VOL
pri .DIR1 ( cfa -- )
	' DODIR W!
---	always use FILE#0 for directory and any temp read only
	0 FILE FCLOSE
---     scan and list rootdir by sector up to one cluster long (!!! COULD BE LONGER)
	@cwd @ S/C
	ADO
	  I SECTOR
---	  terminate listing if this sector is empty
	  SDBUF $80 DATA?
\       	0 SDBUF $200 ADO I @ OR DUP IF LEAVE THEN $20 +LOOP
	  IF
---	    scan the buffer for dir entries --- up to 16 dir entries/sector
	    SDBUF BLKSIZ
	    ADO
---	      accept only valid entries
	      @ATR I+ C@ $0F >   IC@ $80 < AND   IC@ '?' <> AND
---	      all good, list this one using the selected method
	        IF DODIR THEN
---	    next dir entry 32 bytes (assumes standard 8.3)
	    $20 +LOOP
---	    no more dir sectors - leave loop
	  ELSE LEAVE
	  THEN
	LOOP
	;


{
TAQOZ# SEE DIR
$3034: $014F     $2FCA (12234) ' '
$3038: $2A64     ?MOUNT
$303A: $1026     CRLF
$303C: $2F30     pwd
$303E: $2966     .VOL
$3040: $014F     $3030 (12336) '0'
$3044: $00F3     W!
$3046: $FE00     #0
$3048: $2800     FILE
$304A: $2D9E     FCLOSE
$304C: $2892     @cwd
$304E: $00EB     @
$3050: $28FA     S/C
$3052: $0117     ADO
$3054: $010F     I
$3056: $272A     SECTOR
$3058: $24DA     SDBUF
$305A: $FE80     ' ' #128 $0080
$305C: $0109     DATA?
$305E: $FC15     IF $308A
$3060: $24DA       SDBUF
$3062: $2486       BLKSIZ
$3064: $0117       ADO
$3066: $FE0B       #11 $000B
$3068: $0115       I+
$306A: $00E7       C@
$306C: $FE0F       #15 $000F
$306E: $0CCC       >
$3070: $0113       IC@
$3072: $FE80       ' ' #128 $0080
$3074: $0CD4       <
$3076: $008E       AND
$3078: $0113       IC@
$307A: $FE3F       '?' #63 $003F
$307C: $00C1       <>
$307E: $008E       AND
$3080: $FC01       IF $3084
$3082: $3030         DODIR
                   THEN
$3084: $FE20       ' ' #32 $0020
$3086: $0120       +LOOP
$3088: $308D       GOTO $308C
                 THEN
$308A: $0136     LEAVE
$308C: $0123     LOOP
$308E: $005A     ;  ok
}
---         CE1372  .JPG .....a 0000.7C17    20/06/2013 02:36:52   736,486
---  CR ." NAME    .EXT ATRS   SECTOR       DATE       TIME       SIZE "

byte lscnt

pri (ls)
---	5 files per line
	lscnt C@ 5 // 0= IF CRLF THEN
---	format a directory name
	$10 @ATR I+ SET?
	IF
---	  Use [ ] to enclose dir name
	  I 8 .[	\ '[' EMIT I 8 CTYPE ']' EMIT
	  5
	ELSE
---	  normal, print name and opt ext
	  I .F83 3
	THEN
	SPACES lscnt C++
	;


pri .[]?	$10 @ATR I+ SET? AND EMIT ;;

--- List a single directory entry in FTP compatible format
pri (LIST) ( <index> )
---	dummy permissions listing    TO DO !!!
	." -rwxrwxrwx 1 502     500"
	I .FSIZE SPACE @FDATE I+ W@ .FDATE
---	most systems expect either the time or the year if it is older than 6 months
	I @FTIME + W@ .FTIME
---	Print the file name at the current loop index I
\ pri .FNAME ( --- )
---	skip invalid index/entry
	IC@ $20 > IF '[' .[]? I .F83 ']' .[]? THEN
	CRLF
	;;


--- used by FTP

pub LIST        ' (LIST) .DIR1 ;;				--- set list method

--- list the directory
--- ls -l in long format
--- ls -f list open files

pre ls
     delim 1+ C@ $20 =
     IF
       GET$ 1+ C@ SWITCH
       'l' CASE CRLF LIST BREAK
       'f' CASE FILE# 4 0 DO CRLF I .AS" #|# " I FILE @DIRBUF .DIR LOOP FILE BREAK
     THEN
     lscnt C~ ' (ls) (DIR)
    ;;



{ HELP QV ( <filename> -- )
Quick view of file header in ASCII dump format
QV<cr> will display current file if opened
}
pre QV  delim 1+ C@ $20 = IF GET$ DUPC@ IF FOPEN$ THEN DROP THEN FSECT@ 0EXIT 0 $400 FS DUMPA ;; [

{
--- backup memory into P2BOOT.IMG for the size specified in k and setup sector 0
--- takes around 7 seconds on CV-A9 at 80MHz for 1MB image ~= 3ms/blk or 0.5sec for 64k
pub SAVE ( kbytes -- )						--- size in kilobytes where 128 = 128k
	" P2BOOT.IMG" FOPEN$					--- try to open default boot image but exit if failed
	IF
	  1 MAX 1024 MIN 10 <<					--- limit between 1k and 1M and shift up kbytes to bytes
	  FSECT@ 0 3RD						--- use raw file's starting sector from start for size
	    ADO I OVER SDWR DROP 1+ SPIN BLKSIZ +LOOP DROP	--- write the image with a little spinning progress symbol | / - \
	  @DIRBUF 0 RDSDA $20 CMOVE				--- make a copy of the directory entry for reference and detection in sector 0
	  FSECT@ $20 SD!                      			--- Setup a pointer to the raw file starting sector
	  DUP $24 SD!						--- write the file size we are using
	  diradr @FILE @ $28 SD!					--- save 64GB encoded pointer to the directory entry
\	  TIME@ $30 SD!						--- time and date stamp
\	  DATE@ $34 SD!
	  TRUE FLUSH                                          	--- write the sector to the card
	  SWAP .DECL ."  bytes written"		--- Confirmation
	THEN
	DROP							--- discard size or wrsect flag
	;
}



END





{
180124 12548
180122
CODE	13114
NAMES	4142 (527)


180121
HERE . 13172 ok
TAQOZ# @NAMES NFA' END 7 + SWAP - . 4150 ok

17322 TOTAL
527 NAMES

4150 - 527 = 3623 if atrs are combined with count
2042 bytes & characters
3 chars/names average


3167 bytes dictionary of 5-bit long fixed name used
3623 if atrs are simply combined with count

}

